# SymSpell algorithm
# Taken from https://github.com/ne3x7/pysymspell/ as of 2018-07-24
# and modified to make the input format match skills-ml's expectations more closely
# SymSpell original copyright:
#
# MIT License
#
# Copyright (c) 2018 Nikolay Stulov (Python Port https://github.com/ne3x7/pysymspell)
# Copyright (c) 2018 Hampus Londögård (Java Port https://github.com/Lundez/JavaSymSpell)
# Copyright (c) 2018 Wolf Garbe (Original C# implementation https://github.com/wolfgarbe/SymSpell)
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
# https://opensource.org/licenses/MIT



import numpy as np
import os
import sys

class SymSpell():
    """SymSpell: 1 million times faster through Symmetric Delete spelling correction algorithm.
    The Symmetric Delete spelling correction algorithm reduces the complexity of edit candidate generation and dictionary lookup
    for a given Damerau-Levenshtein distance. It is six orders of magnitude faster and language independent.
    Opposite to other algorithms only deletes are required, no transposes + replaces + inserts.
    Transposes + replaces + inserts of the input term are transformed into deletes of the dictionary term.
    Replaces and inserts are expensive and language dependent: e.g. Chinese has 70,000 Unicode Han characters!
    SymSpell supports compound splitting / decompounding of multi-word input strings with three cases:
    1. mistakenly inserted space into a correct word led to two incorrect terms 
    2. mistakenly omitted space between two correct words led to one incorrect combined term
    3. multiple independent input terms with/without spelling errors
    See https://github.com/wolfgarbe/SymSpell for details.
    Args:
        max_dictionary_edit_distance (int, optional): Maximum distance used to generate index. Also acts 
            as an upper bound for `max_edit_distance` parameter in `lookup()` method. Defaults to 2.
        prefix_length (int, optional): Prefix length. Should not be changed normally. Defaults to 7.
        count_threshold (int, optional): Threshold corpus-count value for words to be considered correct.
            Defaults to 1, values below zero are also mapped to 1. Consider setting a higher value if your
            corpus contains mistakes.
    """
    def __init__(self, max_dictionary_edit_distance=2, prefix_length=7, count_threshold=1):
        self._distance_algorithm = 'damerau'
        self._max_length = 0
        self._deletes = None
        self._words = None
        self._below_threshold_words = dict()

        if max_dictionary_edit_distance < 0:
            max_dictionary_edit_distance = 2
        if prefix_length < 1 or prefix_length <= max_dictionary_edit_distance:
            prefix_length = 7
        if count_threshold < 0:
            count_threshold = 1

        self._words = dict()
        self._max_dictionary_edit_distance = max_dictionary_edit_distance
        self._prefix_length = prefix_length
        self._count_threshold = count_threshold
        self._compact_mask = (0xffffffff >> 8) << 2

    def _create_dictionary_entry(self, key, count):
        """Creates or updates a dictionary entry.
        Args:
            key (str): Word to insert or update.
            count (int): Count to save or add to existing.
        Returns:
            bool: True if word was added to the dictionary, False if word was updated or ignored.
        """
        if count <= 0:
            if self._count_threshold > 0:
                return False
            count = 0

        if self._count_threshold > 1 and key in self._below_threshold_words.keys():
            count_previous = self._below_threshold_words[key]
            count = count_previous + count
            if count >= self._count_threshold:
                self._below_threshold_words.pop(key)
            else:
                self._below_threshold_words[key] = count
                return False
        elif key in self._words.keys():
            count_previous = self._words[key]
            count = count_previous + count
            self._words[key] = count
            return False
        elif count < self._count_threshold:
            self._below_threshold_words[key] = count
            return False

        self._words[key] = count
        if len(key) > self._max_length:
            self._max_length = len(key)

        edits = self._edits_prefix(key)

        if self._deletes is None:
            self._deletes = dict()

        for edit in edits:
            hs = self._hash(edit)
            suggestions = list()
            if hs in self._deletes.keys():
                suggestions = self._deletes.get(hs)
                suggestions.append(key)
                self._deletes[hs] = suggestions
            else:
                suggestions = [key]
                self._deletes[hs] = suggestions
        return True

    def load_dictionary(self, corpus):
        """Loads dictionary from :param:`corpus` file.
        File should contain space-separated word-count pairs one at a line.
        Args:
            corpus (str): Path to corpus file.
        """
        if os.path.exists(corpus):
            with open(corpus, 'r') as f:
                for line in f:
                    key, count = line.split()
                    count = int(count)
                    self._create_dictionary_entry(key, count)

        if self._deletes is None:
            self._deletes = dict()

    def create_dictionary(self, corpus):
        """Creates dictionary from :param:`corpus` file.
        Note:
            Words are not preprocessed in any way. It is your duty to provide appropriate corpus. Also
                keep in mind that the distance used to generate index is specified at initialization. Consider
                doing a purge of below threshold words afterwards.
        Args:
            corpus (str): Path to corpus file.
        """
        for term in corpus:
            self._create_dictionary_entry(term, 1)

        if self._deletes is None:
            self._deletes = dict()

    def purge_below_threshold_words(self):
        """Purges words below threshold.
        Consider using this method after creating a dictionary to reduce memory usage. These words are not
            used in any way during lookup.
        """
        self._below_threshold_words = dict()

    def lookup(self, phrase, verbosity, max_edit_distance):
        """Attempts to correct the spelling of :param:`phrase`.
        Note:
            Phrase is not preprocessed in any way.
        Args:
            phrase (str): Word to correct. Should be a valid word.
            verbosity (int, 0, 1 or 2): Output toggle. Set to 0 to output closest most common correction,
                set to 1 to output closest suggestion, set to 2 to output all suggestions.
            max_edit_distance (int): Maximum edit distance to consider.
        Returns:
            list of :obj:`SuggestionItem`: Suggested corrections.
        Raises:
            AssertionError: If :param:`max_edit_distance` is larger than maximum edit distance specified 
                at initialization.
        """
        assert max_edit_distance <= self._max_dictionary_edit_distance, 'Distance too big'
        suggestions = list()
        phrase_len = len(phrase)

        if phrase_len - max_edit_distance > self._max_length:
            return suggestions

        considered_deletes = set()
        considered_suggestions = set()
        suggestions_count = 0

        if phrase in self._words.keys():
            suggestions_count = self._words[phrase]
            suggestions.append(SuggestionItem(phrase, 0, suggestions_count))
            if verbosity < 2:
                return suggestions
        considered_suggestions.add(phrase)

        max_edit_distance_2 = max_edit_distance
        candidate_ptr = 0
        candidates = list()

        phrase_prefix_len = phrase_len
        if phrase_prefix_len > self._prefix_length:
            phrase_prefix_len = self._prefix_length
            candidates.append(phrase[:phrase_prefix_len])
        else:
            candidates.append(phrase)

        comp = EditDistance(phrase, self._distance_algorithm)

        while candidate_ptr < len(candidates):
            candidate = candidates[candidate_ptr]
            candidate_ptr += 1
            candidate_len = len(candidate)
            len_diff = phrase_prefix_len - candidate_len

            if len_diff > max_edit_distance_2:
                if verbosity == 2:
                    continue
                else:
                    break

            if self._hash(candidate) in self._deletes.keys():
                dictionary_suggestions = self._deletes[self._hash(candidate)]
                for suggestion in dictionary_suggestions:
                    if suggestion == phrase:
                        continue
                    suggestion_len = len(suggestion)

                    if abs(suggestion_len - phrase_len) > max_edit_distance_2 \
                            or suggestion_len < candidate_len \
                            or (suggestion_len == candidate_len and suggestion != candidate):
                        continue

                    suggestion_prefix_len = min(suggestion_len, self._prefix_length)
                    if suggestion_prefix_len > phrase_prefix_len and suggestion_prefix_len - candidate_len > max_edit_distance_2:
                        continue

                    distance = np.inf
                    min_distance = 0

                    if candidate_len == 0:
                        distance = max(phrase_len, suggestion_len)
                        if distance > max_edit_distance_2 or suggestion in considered_suggestions:
                            considered_suggestions.add(suggestion)
                            continue
                        considered_suggestions.add(suggestion)
                    elif suggestion_len == 1:
                        if phrase.index(suggestion[0]) < 0:
                            distance = phrase_len
                        else:
                            distance = phrase_len - 1
                        if distance > max_edit_distance_2 or suggestion in considered_suggestions:
                            considered_suggestions.add(suggestion)
                            continue
                        considered_suggestions.add(suggestion)
                    else:
                        min_distance = min(phrase_len, suggestion_len) - self._prefix_length
                        if self._prefix_length - max_edit_distance == candidate_len \
                            and (min_distance > 1 and phrase[:phrase_len + 1 - min_distance] != suggestion[:phrase_len + 1 - min_distance]) \
                            or (min_distance > 0 and phrase[phrase_len - min_distance] != suggestion[suggestion_len - min_distance] \
                                and (phrase[phrase_len - min_distance - 1] != suggestion[suggestion_len - min_distance] \
                                or phrase[phrase_len - min_distance] != suggestion[suggestion_len - min_distance - 1])):
                            continue
                        else:
                            if (verbosity < 2 and not self._delete_in_suggestion_prefix(candidate, candidate_len, suggestion, suggestion_len)) \
                                or suggestion in considered_suggestions:
                                considered_suggestions.add(suggestion)
                                continue
                            considered_suggestions.add(suggestion)
                            distance = comp.compare(suggestion, max_edit_distance_2)
                            if distance < 0:
                                continue

                    if distance <= max_edit_distance_2:
                        suggestion_count = self._words[suggestion]
                        si = SuggestionItem(suggestion, distance, suggestion_count)
                        if len(suggestions) > 0:
                            if verbosity == 0: # Top
                                if distance < max_edit_distance_2 or suggestion_count > suggestions[0].count:
                                    max_edit_distance_2 = distance
                                    suggestions[0] = si
                                continue
                            elif verbosity == 1: # Closest
                                if distance < max_edit_distance_2:
                                    suggestions = list()
                                break
                        if verbosity < 2:
                            max_edit_distance_2 = distance
                        suggestions.append(si)

            if len_diff < max_edit_distance and candidate_len <= self._prefix_length:
                if verbosity < 2 and len_diff >= max_edit_distance_2:
                    continue

                for i in range(candidate_len):
                    d = candidate[:i] + candidate[i+1:]

                    if d not in considered_deletes:
                        considered_deletes.add(d)
                        candidates.append(d)

        if len(suggestions) > 1:
            suggestions.sort()
        return suggestions

    def lookup_compound(self, phrase, max_edit_distance):
        """Attempts to correct the spelling of :param:`phrase`.
        Note:
            Phrase is not preprocessed in any way.
        Args:
            phrase (str): Sentence to correct.
            max_edit_distance (int): Maximum edit distance to consider for each word.
        Returns:
            list of :obj:`SuggestionItem`: Length-one list with suggested correction.
        Raises:
            AssertionError: If :param:`max_edit_distance` is larger than maximum edit distance specified 
                at initialization.
        """
        assert max_edit_distance <= self._max_dictionary_edit_distance, 'Distance too big'

        terms_list_1 = self._parse_words(phrase)
        suggestions = list()
        suggestions_parts = list()
        suggestions_combi = list()
        edit_distance = np.inf

        last_combi = False
        for i in range(len(terms_list_1)):
            suggestions = self.lookup(terms_list_1[i], 0, max_edit_distance)
            if i > 0 and not last_combi:
                suggestions_combi = self.lookup(terms_list_1[i-1] + terms_list_1[i], 0, max_edit_distance)
                if len(suggestions_combi) > 0:
                    best_1 = suggestions_parts[-1]
                    if len(suggestions) > 0:
                        best_2 = suggestions[0]
                    else:
                        best_2 = SuggestionItem(terms_list_1[i], max_edit_distance+1, 0)

                    distance = EditDistance(terms_list_1[i-1] + ' ' + terms_list_1[i], 'damerau')
                    if suggestions_combi[0].distance + 1 < distance.damerau_levenshtein_distance(best_1.term + ' ' + best_2.term, max_edit_distance):
                        suggestions_combi[0].distance += 1
                        suggestions_parts[-1] = suggestions_combi[0]
                        last_combi = True
                        continue

            last_combi = False

            if len(suggestions) > 0 and (suggestions[0].distance == 0 or len(terms_list_1[i]) == 1):
                suggestions_parts.append(suggestions[0])
            else:
                suggestions_split = list()

                if len(suggestions) > 0:
                    suggestions_split.append(suggestions[0])

                if len(terms_list_1[i]) > 1:
                    for j in range(1, len(terms_list_1[i])):
                        part_1 = terms_list_1[i][:j]
                        part_2 = terms_list_1[i][j:]
                        suggestions_1 = self.lookup(part_1, 1, max_edit_distance)

                        if len(suggestions_1) > 0:
                            if len(suggestions) > 0 and suggestions[0] == suggestions_1[0]:
                                continue
                            suggestions_2 = self.lookup(part_2, 1, max_edit_distance)

                            if len(suggestions_2) > 0:
                                if len(suggestions) > 0 and suggestions[0] == suggestions_2[0]:
                                    continue

                                split = suggestions_1[0].term + ' ' + suggestions_2[0].term
                                edit_distance = EditDistance(terms_list_1[i], 'damerau')
                                suggestion_split = SuggestionItem(split, 
                                    edit_distance.damerau_levenshtein_distance(split, max_edit_distance), 
                                    min(len(suggestions_1), len(suggestions_2)))
                                if suggestion_split.distance >= 0:
                                    suggestions_split.append(suggestion_split)
                                if suggestion_split.distance == 1:
                                    break

                    if len(suggestions_split) > 0:
                        suggestions_split.sort()
                        suggestions_parts.append(suggestions_split[0])
                    else:
                        si = SuggestionItem(terms_list_1[i], 0, max_edit_distance + 1)
                        suggestions_parts.append(si)

                else:
                    si = SuggestionItem(terms_list_1[i], 0, max_edit_distance + 1)
                    suggestions_parts.append(si)

        suggestion = SuggestionItem('', sys.maxsize, sys.maxsize)
        s = ' '.join([x.term for x in suggestions_parts])
        suggestion.count = min([x.count for x in suggestions_parts])
        suggestion.term = s
        edit_distance = EditDistance(suggestion.term, 'damerau')
        suggestion.distance = edit_distance.damerau_levenshtein_distance(phrase, self._max_dictionary_edit_distance)
        return [suggestion]

    def _delete_in_suggestion_prefix(self, delete, delete_len, suggestion, suggestion_len):
        """Helper method to check if :param:`delete` is prefix of :param:`suggestion`.
        Args:
            delete (str): String to look for in prefix.
            delete_len (int): Length of :param:`delete`.
            suggestion (str): String to take prefix from.
            suggestion_len (int): Length of :param:`suggestion`.
        Returns:
            bool: True if :param:`delete` is prefix of :param:`suggestion`, False otherwise.
        """
        if delete_len == 0:
            return True
        if self._prefix_length < suggestion_len:
            suggestion_len = self._prefix_length

        j = 0
        for i in range(delete_len):
            ch = delete[i]

            while j < suggestion_len and ch != suggestion[j]:
                j += 1

            if j == suggestion_len:
                return False
        return True

    def _edits(self, word, edit_distance, delete_words):
        """helper recursive method to generate deletes.
        Refer to article for details.
        Args:
            word (str): Word to generate deletes from.
            edit_distance (int): Maximum edit distance to consider, recursion depth.
            delete_words (set): Generated deletes, pass empty set first time.
        Returns:
            delete_words (set): Generated deletes.
        """
        edit_distance += 1
        for i in range(len(word)):
            delete = word[:i] + word[i+1:]
            if delete not in delete_words:
                delete_words.add(delete)
                if edit_distance < self._max_dictionary_edit_distance:
                    self._edits(delete, edit_distance, delete_words)
        return delete_words

    def _edits_prefix(self, key):
        s = set()
        if len(key) <= self._max_dictionary_edit_distance:
            s.add('')
        if len(key) > self._prefix_length:
            key = key[:self._prefix_length]
        s.add(key)
        return self._edits(key, 0, s)

    def _hash(self, s):
        l = len(s)
        l_mask = l
        if l_mask > 3:
            l_mask = 3

        hs = 2166136261
        for i in range(l):
            hs ^= ord(s[i])
            hs *= 16777619
        hs &= self._compact_mask
        hs |= l_mask
        return hs

    def _parse_words(self, text, filters='!"#$%&()*+,-./:;<=>?@[\\]^_`{|}~\t\n', lower=True, split=' '):
        if lower:
            text = text.lower()

        translate_map = str.maketrans(filters, split * len(filters))

        text = text.translate(translate_map)
        seq = text.split(split)
        return [i for i in seq if i]

class EditDistance():
    def __init__(self, base_string, distance_algorithm):
        self._base_string = base_string
        self._distance_algorithm = distance_algorithm
        if self._base_string == '':
            self._base_string = None
            return
        if self._distance_algorithm == 'damerau':
            self._v0 = np.zeros(len(self._base_string), dtype=np.int32)
            self._v2 = np.zeros(len(self._base_string), dtype=np.int32)

    def compare(self, string_2, max_distance):
        if self._distance_algorithm == 'damerau':
            return self.damerau_levenshtein_distance(string_2, max_distance)

    def damerau_levenshtein_distance(self, string_2, max_distance):
        if self._base_string is None:
            if string_2 is None:
                return 0
            else:
                return len(string_2)

        if string_2 is None or string_2 == '':
            return len(self._base_string)

        if len(self._base_string) > len(string_2):
            string_1 = string_2
            string_2 = self._base_string
        else:
            string_1 = self._base_string

        slen = len(string_1)
        tlen = len(string_2)

        while slen > 0 and string_1[slen-1] == string_2[tlen-1]:
            slen -= 1
            tlen -= 1

        st = 0
        if string_1[0] == string_2[0] or slen == 0:
            while st < slen and string_1[st] == string_2[st]:
                st += 1

            slen -= st
            tlen -= st

            if slen == 0:
                return tlen

            string_2 = string_2[st:st+tlen]

        len_diff = tlen - slen
        if max_distance < 0 or max_distance > tlen:
            max_distance = tlen
        elif len_diff > max_distance:
            return -1

        if tlen > self._v0.shape[0]:
            self._v0 = np.zeros(tlen, dtype=np.int32)
            self._v2 = np.zeros(tlen, dtype=np.int32)
        else:
            for i in range(tlen):
                self._v2[i] = 0

        for j in range(max_distance):
            self._v0[j] = j + 1

        for j in range(max_distance, tlen):
            self._v0[j] = max_distance + 1

        j_st_offset = max_distance - (tlen - slen)
        have_max = max_distance < tlen
        j_st = 0
        j_fn = max_distance
        s_ch = string_1[0]
        cur = 0
        for i in range(slen):
            prev_s_ch = s_ch
            s_ch = string_1[st + i]
            t_ch = string_2[0]
            left = i
            cur = left + 1
            next_trans_cost = 0
            if i > j_st_offset:
                j_st += 1
            if j_fn < tlen:
                j_fn += 1
            for j in range(j_st, j_fn):
                above = cur
                this_trans_cost = next_trans_cost
                next_trans_cost = self._v2[j]
                cur = left
                self._v2[j] = cur
                left = self._v0[j]
                prev_t_ch = t_ch
                t_ch = string_2[j]
                if s_ch != t_ch:
                    if left < cur:
                        cur = left
                    if above < cur:
                        cur = above
                    cur += 1
                    if i != 0 and j != 0 and s_ch == prev_t_ch and prev_s_ch == t_ch:
                        this_trans_cost += 1
                        if this_trans_cost < cur:
                            cur = this_trans_cost
                self._v0[j] = cur
            if have_max and self._v0[i + len_diff] > max_distance:
                return -1
        if cur <= max_distance:
            return cur
        else:
            return -1

class SuggestionItem():
    def __init__(self, term, distance, count):
        self._term = term
        self._distance = distance
        self._count = count

    def __eq__(self, other):
        if self._distance == other._distance:
            return self._count == other._count
        else:
            return self._distance == other._distance

    def __lt__(self, other):
        if self._distance == other._distance:
            return self._count > other._count
        else:
            return self._distance < other._distance

    def __str__(self):
        return self._term + ':' + str(self._count) + ':' + str(self._distance)

    def __repr__(self):
        return self._term + ':' + str(self._count) + ':' + str(self._distance)

    @property
    def count(self):
        return self._count

    @count.setter
    def count(self, count):
        self._count = count

    @property
    def distance(self):
        return self._distance

    @distance.setter
    def distance(self, distance):
        self._distance = distance

    @property
    def term(self):
        return self._term

    @term.setter
    def term(self, term):
        self._term = term



from torcherist.algorithms.nlp import clean_html, lowercase_strip_punc

class CorpusCreator(object):
    """
        A base class for objects that convert common schema
        job listings into a corpus in documnet level suitable for use by
        machine learning algorithms or specific tasks.

    Example:
    ```python
    from skills_ml.job_postings.common_schema import JobPostingCollectionSample
    from skills_ml.job_postings.corpora.basic import CorpusCreator

    job_postings_generator = JobPostingCollectionSample()

    # Default will include all the cleaned job postings
    corpus = CorpusCreator(job_postings_generator)

    # For getting a the raw job postings without any cleaning
    corpus = CorpusCreator(job_postings_generator, raw=True)
    ```


    Attributes:
        job_posting_generator (generator):  an iterable that generates JSON strings.
                                Each string is expected to represent a job listing
                                conforming to the common schema
                                See sample_job_listing.json for an example of this schema
        document_schema_fields (list): an list of schema fields to be included
        raw (bool): a flag whether to return the raw documents or transformed documents

    Yield:
        (dict): a dictinary only with selected fields as keys and corresponding raw/cleaned value
    """
    def __init__(self, job_posting_generator=None, document_schema_fields=['description','experienceRequirements', 'qualifications', 'skills'],
                 raw=False):
        self.job_posting_generator = job_posting_generator
        self.raw = raw
        self.document_schema_fields = document_schema_fields
        self.join_spaces = ' '.join
        self.key = ['onet_soc_code']


    @property
    def metadata(self):
        meta_dict = {'corpus_creator': ".".join([self.__module__ , self.__class__.__name__])}
        if self.job_posting_generator:
            meta_dict.update(self.job_posting_generator.metadata)
        return meta_dict

    def _clean(self, document):
        for f in self.document_schema_fields:
            try:
                cleaned = clean_html(document[f]).replace('\n','')
                cleaned = " ".join(cleaned.split())
                document[f] = cleaned
            except KeyError:
                pass
        return document

    def _transform(self, document):
        if self.raw:
            return self._join(document)
        else:
            return self._clean(document)

    def _join(self, document):
        return self.join_spaces([
            document.get(field, '') for field in self.document_schema_fields
        ])

    def __iter__(self):
        for document in self.job_posting_generator:
            document = {key: document[key] for key in self.document_schema_fields}
            yield self._transform(document)


class SimpleCorpusCreator(CorpusCreator):
    """
        An object that transforms job listing documents by picking
        important schema fields and returns them as one large lowercased string
    """
    def _clean(self, document):
        return self.join_spaces([
            lowercase_strip_punc(document.get(field, ''))
            for field in self.document_schema_fields
        ])



import io
import csv

import logging
from cached_property import cached_property
import requests

from collections.abc import MutableMapping, KeysView

from . import CompetencyOntology, Competency, Occupation
from torcherist.config import dataset_dir
from pathlib import Path

logger = logging.getLogger(__name__)
logger.addHandler(logging.StreamHandler())


class KeysViewOnlyKeys(KeysView):
    # Subclass that just changes the representation.
    def __repr__(self):
        return f"{self.__class__.__name__}({list(self)})"


class Clustering(MutableMapping):
    """ A clustering object acting like a dictionary which key is a cluster concept
    and value is a list of entities associated to that cluster.

    Note:
        Python allows a key to be custom object and not necessarily to be string or
        integer as long as it's hashable, but an object key can be difficult to access.
        `key_transform_fn` is to transform an object key to something else, like string
        or integer for easier accessing clustering.

        `value_item_transfrom_fn` is to convert an abstract entity object into something
        else like string, or integer for further computation. It could be a function to
        concat several attributes of the object.

    Example:
        To create a clustering object that we will iterate through a series of concept and
        entity objects, and build the whole thing, we want to extract the concept name
        attribute as the key and make a tuple of entity's identifier and name as the value.
        ```python
        d = Clustering(
                    name="major_group_competencies_name",
                    key_transform_fn=lambda concept: getattr(concept, "name"),
                    value_item_transform_fn=lambda entity: (getattr(entity, "identifier"), getattr(entity, "name")),
            )
     ```

    Args:
        name (str): Name of the clustering
        key_transform_fn (func): the transform function for keys
        value_item_transform_fn (func): the transform function for values

    """

    def __init__(self,
                 name,
                 key_transform_fn=lambda concept: concept,
                 value_item_transform_fn=lambda entity: entity, ):
        self.name = name
        self.store = dict()
        self.map_raw_key = dict()
        self.map_raw_value = dict()
        self.key_transform_fn = key_transform_fn
        self.value_item_transform_fn = value_item_transform_fn

    def __setitem__(self, key, value):
        self.store[self.__keytransform__(key)] = [self.value_item_transform_fn(v) for v in value]
        self.map_raw_value[self.__keytransform__(key)] = value
        self.map_raw_key[self.__keytransform__(key)] = key

    def __getitem__(self, key):
        return self.store[key]

    def __delitem__(self, key):
        del self.store[key]
        del self.map_raw_key[key]
        del self.map_raw_value[key]

    def __iter__(self):
        return iter(self.store)

    def __len__(self):
        return len(self.store)

    def __keytransform__(self, key):
        return self.key_transform_fn(key)

    def keys(self):
        return KeysViewOnlyKeys(self)

    def raw_items(self):
        return self.map_raw_value.items()


majorgroupname = {
    '11': 'Management',
    '13': 'Business and Financial Operations',
    '15': 'Computer and Mathematical',
    '17': 'Architecture and Engineering',
    '19': 'Life, Physical, and Social Science',
    '21': 'Community and Social Service',
    '23': 'Legal',
    '25': 'Education, Training, and Library',
    '27': 'Arts, Design, Entertainment, Sports, and Media',
    '29': 'Healthcare Practitioners and Technical',
    '31': 'Healthcare Support',
    '33': 'Protective Service',
    '35': 'Food Preparation and Serving Related',
    '37': 'Building and Grounds Cleaning and Maintenance',
    '39': 'Personal Care and Service',
    '41': 'Sales and Related',
    '43': 'Office and Administrative Support',
    '45': 'Farming, Fishing, and Forestry',
    '47': 'Construction and Extraction',
    '49': 'Installation, Maintenance, and Repair',
    '51': 'Production',
    '53': 'Transportation and Material Moving',
    '55': 'Military Specific'
}


class OnetReader:
    """
    An object that downloads files from the ONET site
    """

    def __init__(self, onet_dir: Path, version="25_1"):
        self.onet_dir = onet_dir
        self.onet_dir.mkdir(parents=True, exist_ok=True)
        self.downloader = OnetDownloader(to=self.onet_dir, version=version)

    def reader(self, filename: str) -> csv.DictReader:
        """
        Ensures that the given ONET data file is present, either by
        using a cached copy or downloading from S3

        Args:
            filename: unpathed filename of an ONET file (eg. Skills, Tools)

        Returns:
            csv.DictReader
        """
        target_file = self.onet_dir / filename
        if not target_file.exists():
            # download & save to onet_dir/filename
            self.downloader.download(filename)
        # now onet_dir/filename exists
        return csv.DictReader(target_file.open("r"), delimiter="\t")


class OnetDownloader(object):
    """
    Downloads `version` of ONET
     and save to `to` dir
    """

    def __init__(self, to: Path, version: str):
        self.to = to
        self.to.mkdir(exist_ok=True, parents=True)
        self.version = version
        self.url_prefix = f'http://www.onetcenter.org/dl_files/database/db_{version}_text'

    def download(self, source_file):
        url = f'{self.url_prefix}/{source_file}.txt'
        response = requests.get(url)
        with open(self.to / source_file, "w") as f:
            f.write(response.text)


class Onet(CompetencyOntology):
    def __init__(self, version="25_1", onet_dir=None):
        super().__init__()
        self.reader = OnetReader(onet_dir=onet_dir or Path(dataset_dir) / "onet",
                                 version=version)
        self.name = 'onet'
        self.competency_framework.name = 'onet_ksat'
        self.competency_framework.description = 'ONET Knowledge, Skills, Abilities, Tools, and Technology'
        self._build()

    def _build(self):
        reader = self.reader
        description_lookup = {}
        logging.info('Processing Content Model Reference')
        for row in reader.reader('Content Model Reference'):
            description_lookup[row['Element ID']] = row['Description']

        logging.info('Processing occupation data')
        for row in reader.reader('Occupation Data'):
            occupation = Occupation(
                identifier=row['O*NET-SOC Code'],
                name=row['Title'],
                description=row['Description'],
                categories=['O*NET-SOC Occupation'],
            )
            major_group_num = row['O*NET-SOC Code'][0:2]
            major_group = Occupation(
                identifier=major_group_num,
                name=majorgroupname[major_group_num],
                categories=['O*NET-SOC Major Group']
            )
            occupation.add_parent(major_group)
            self.add_occupation(occupation)
            self.add_occupation(major_group)

        for content_model_file in [
            'Knowledge', 'Abilities', 'Skills', 'Interests',
            'Work Styles', 'Work Context', 'Work Activities',
        ]:
            logging.info(f"Processing {content_model_file}")
            for row in reader.reader(content_model_file):
                scale, score = row["Scale ID"], float(row["Data Value"])
                # for knowledge, AbilitSkills, WorkStyle, WorkActivity: importance IM (1-5)
                # for Interests: RIASEC level OI of interest (1-7)
                # for Work Context: percent of respondents endorsing CXP (0-100)
                if scale in ['IM', 'OI', 'CXP']:
                    competency = Competency(
                        identifier=row['Element ID'],
                        name=row['Element Name'],
                        categories=[content_model_file, f"{scale}-{score}"],
                        competencyText=description_lookup[row['Element ID']]
                    )
                    self.add_competency(competency)
                    occupation = Occupation(identifier=row['O*NET-SOC Code'])
                    self.add_edge(competency=competency, occupation=occupation)

        logging.info('Processing tools and technology')
        for file in ["Technology Skills", "Tools Used"]:
            for row in reader.reader(file):
                commodity = row['Commodity Title']
                cat = "Tech Skills" if file == "Technology Skills" else "Tools"
                hot = ["HOT"] if file == "Technology Skills" and row["Hot Technology"] == "Y" else []
                # no competencyText
                competency = Competency(
                    # id = <commodity id>-<skill name> since otherwise only add commodity
                    identifier=row["Commodity Code"] + "-" + row["Example"],
                    name=row['Example'],
                    categories=[cat] + hot + [commodity],
                )
                self.add_competency(competency)
                occupation = Occupation(identifier=row['O*NET-SOC Code'])
                self.add_edge(competency=competency, occupation=occupation)

        logging.info("Processing tasks")
        for row in reader.reader("Task Statements"):
            competency = Competency(
                identifier=row['Task ID'],
                name=row['Task'],
                categories=["Tasks"] + [row["Task Type"]],
            )
            self.add_competency(competency)
            occupation = Occupation(identifier=row['O*NET-SOC Code'])
            self.add_edge(competency=competency, occupation=occupation)

    @cached_property
    def all_soc(self):
        occupations = self.occupations
        soc = []
        for occ in occupations:
            if 'O*NET-SOC Occupation' in occ.other_attributes['categories']:
                soc.append(occ.identifier)
        return sorted(soc)

    @cached_property
    def all_major_groups(self):
        occupations = self.occupations
        major_groups = []
        for occ in occupations:
            if 'O*NET-SOC Major Group' in occ.other_attributes['categories']:
                major_groups.append(occ)
        return sorted(major_groups, key=lambda k: k.identifier)

    @cached_property
    def all_major_groups_occ(self):
        occ = self.filter_by(lambda edge: len(edge.occupation.identifier) == 2)
        return occ.occupations

    @cached_property
    def competency_categories(self):
        return set(c.categories[0] for c in self.competencies)

    @cached_property
    def major_group_occupation_name_clustering(self):
        d = Clustering(
            name="major_group_occupations_name",
            key_transform_fn=lambda concept: getattr(concept, "name"),
            value_item_transform_fn=lambda entity: (getattr(entity, "identifier"), getattr(entity, "name")),
        )
        for mg in self.all_major_groups_occ:
            d[mg] = [child for child in mg.children]
        return d

    @cached_property
    def major_group_occupation_description_clustering(self):
        d = Clustering(
            name="major_group_occupations_description",
            key_transform_fn=lambda concept: getattr(concept, "name"),
            value_item_transform_fn=lambda entity: (
                getattr(entity, "identifier"),
                ' '.join([str(getattr(entity, "name")), str(getattr(entity, "other_attributes").get("description"))])),
        )
        for mg in self.all_major_groups_occ:
            d[mg] = [child for child in mg.children]
        return d

    @cached_property
    def major_group_competencies_name_clustering(self):
        d = Clustering(
            name="major_group_competencies_name",
            key_transform_fn=lambda concept: getattr(concept, "name"),
            value_item_transform_fn=lambda entity: (getattr(entity, "identifier"), getattr(entity, "name")),
        )
        for mg in self.all_major_groups_occ:
            d[mg] = self.filter_by(lambda edge: edge.occupation.identifier[:2] == mg.identifier[:2]).competencies
        return d

    @cached_property
    def major_group_competencies_description_clustering(self):
        d = Clustering(
            name="major_group_competencies_description",
            key_transform_fn=lambda concept: getattr(concept, "name"),
            value_item_transform_fn=lambda entity: (
                getattr(entity, "identifier"),
                ' '.join(
                    [str(getattr(entity, "name")), str(getattr(entity, "other_attributes").get("competencyText"))])),
        )
        for mg in self.all_major_groups_occ:
            d[mg] = self.filter_by(lambda edge: edge.occupation.identifier[:2] == mg.identifier[:2]).competencies
        return d

    def generate_clusterings(self):
        return [
            self.major_group_occupation_name_clustering,
            self.major_group_occupation_description_clustering,
            self.major_group_competencies_name_clustering,
            self.major_group_competencies_description_clustering
        ]



import torch
import torch.nn as nn
from typing import Optional


def sequence_mask(valid_len, max_len: Optional[int] = None):
    """
    find valid_len's corresponding mask
    Args:
        valid_len: (d_1, ..., d_n), each ele is valid len for some matrix of shape (d_1, ..., d_n, d),
            eg. if valid_len[1,2,3] = 6, then matrix[1,2,3] (an array) after 6th ele would marked irrelevant, thus False
        max_len: if None `max_len = max(valid_len)` the max of all valid_len ele

    Returns:
        mask: (d_1, ..., d_n, max_len),
            mask[i_1, ..., i_n, i] = i < valid_len[i_1, ..., i_n]
            thus matrix[~mask] would be those irrelevant fields
    """
    if not max_len:
        max_len = torch.max(valid_len)
    # (1, m) < (d_1, ..., d_n , 1)
    # broadcast (d_1, ..., d_n, m)
    # LHS each row (high-dim) copy of 1 ... m
    # RHS each column copy of valid_len
    # True if less than valid_len's value, thus relevant
    mask = torch.arange(max_len,
                        dtype=torch.float32,
                        device=valid_len.device)[None, :] < valid_len[..., None]

    return mask

def fill_sequence_mask(X, valid_len, value: int = 0):
    """
    fill mask of X to be `value`
    Args:
        X: input tensor (d_1, ..., d_n, d)
        valid_len: (d_1, ..., d_n), valid_len[i_1, ..., i_n] (int) is the valid len of X[i_1, ..., i_n] (row)
        value: scalar

    Returns:
        None, inplace change
    """
    mask = sequence_mask(valid_len, X.size(-1))
    X[~mask] = value




"""Extract skills from text corpora, such as job postings"""
from .fuzzy_match import FuzzyMatchSkillExtractor
from .exact_match import ExactMatchSkillExtractor
from .soc_exact import SocScopedExactMatchSkillExtractor
from .noun_phrase_ending import SkillEndingPatternExtractor, AbilityEndingPatternExtractor
from .section_extract import SectionExtractSkillExtractor


__all__ = [
    'ExactMatchSkillExtractor',
    'FuzzyMatchSkillExtractor',
    'SocScopedExactMatchSkillExtractor',
    'SkillEndingPatternExtractor',
    'AbilityEndingPatternExtractor',
    'SectionExtractSkillExtractor'
]



"""Use fuzzy matching with a source list to extract skills
from unstructured text"""
import logging
# from descriptors import cachedproperty
from math import ceil

import nltk
try:
    nltk.sent_tokenize('test')
except LookupError:
    nltk.download('punkt')

from .base import (
    CandidateSkill,
    ListBasedSkillExtractor,
    CandidateSkillYielder
)
from typing import Dict, Text, Generator
from .symspell import SymSpell


class FuzzyMatchSkillExtractor(ListBasedSkillExtractor):
    """Extract skills from unstructured text using fuzzy matching"""

    match_threshold = 88
    max_distance = 4
    max_ngrams = 5

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        competencies = set(
            competency
            for competency in self.competency_framework.values()
            if competency.name
        )

        self.id_lookup = dict((competency.name.lower(), competency.identifier) for competency in competencies)

        logging.info(
            'Found %s entries for lookup',
            len(competencies)
        )
        self.symspell = SymSpell(max_dictionary_edit_distance=4)
        self.symspell.create_dictionary(list(self.id_lookup.keys()))

    @property
    def method_name(self) -> Text:
        return f'fuzzy_thresh{self.match_threshold}_maxdist{self.max_distance}_maxngram{self.max_ngrams}'

    @property
    def method_description(self) -> Text:
        return f'Fuzzy matching using ratio of most similar substring, ' + \
            f'with a minimum cutoff of {self.match_threshold} percent match'

    def reg_ex(self, s):
        s = s.replace(".", "\.")
        s = s.replace("^", "\^")
        s = s.replace("$", "\$")
        s = s.replace("*", "\*")
        s = s.replace("+", "\+")
        s = s.replace("?", "\?")
        return s

    def ngrams(self, sent: Text, N: int) -> Generator[Text, None, None]:
        """Yield ngrams from sentence

        Args:
            sent (string) A sentence
            N (int) The maximum N-gram length to create. 

        Yields: (string) n-grams in sentence from n=1 up to the given N
        """
        sent_input = self.nlp.word_tokenize(sent)
        for n in range(1, N):
            for i in range(len(sent_input)-n+1):
                yield " ".join(sent_input[i:i+n]).lower()

    def candidate_skills(self, source_object: Dict) -> CandidateSkillYielder:
        document = self.transform_func(source_object)
        sentences = self.nlp.sentence_tokenize(document, include_spans=True)

        for sent in sentences:
            for phrase in self.ngrams(sent.text, self.max_ngrams):
                length_of_phrase = len(phrase)
                max_distance = length_of_phrase - \
                    ceil(length_of_phrase * self.match_threshold/100)
                if max_distance > self.max_distance:
                    max_distance = self.max_distance
                matches = self.symspell.lookup(phrase, 2, max_distance)
                for match in matches:
                    logging.info(
                        'Fuzzy match found: %s corrected to %s in %s',
                        phrase,
                        match.term,
                        sent.text
                    )
                    yield CandidateSkill(
                        skill_name=phrase,
                        matched_skill_identifier=self.id_lookup[match.term],
                        skill_cat=None,
                        confidence=100*(length_of_phrase-match.distance)/length_of_phrase,
                        context=sent.text,
                        start_index=sent.start_index + sent.text.lower().index(phrase),
                        document_id=source_object['id'],
                        document_type=source_object['@type'],
                        source_object=source_object,
                        skill_extractor_name=self.name
                    )



# from skills_ml.ontologies.base import CompetencyOntology
from torcherist.dataset.ontology import CompetencyOntology

from .exact_match import ExactMatchSkillExtractor


class SocScopedExactMatchSkillExtractor(ExactMatchSkillExtractor):
    """Extract skills from unstructured text,
    but only return matches that agree with a known taxonomy
    """
    method_name = 'occscoped_exact_match'
    method_description = 'Exact matching using only the skills known by ONET to be associated with the given SOC code'

    def __init__(self, competency_ontology, *args, **kwargs):
        if not isinstance(competency_ontology, CompetencyOntology):
            raise ValueError('Must pass in a CompetencyOntology object')
        super().__init__(competency_ontology.competency_framework, *args, **kwargs)

        self.competency_ontology = competency_ontology
        self.skill_extractor_by_soc = {}

    def candidate_skills(self, source_object):
        soc_code = source_object.get('onet_soc_code', None)
        if not soc_code:
            return
        if soc_code not in self.skill_extractor_by_soc:
            subontology = self.competency_ontology.filter_by(
                lambda edge: edge.occupation.identifier == soc_code,
                competency_name='competencies_' + soc_code,
                competency_description='Competencies needed by ' + soc_code
            )
            if len(subontology.competencies) == 0:
                return

            self.skill_extractor_by_soc[soc_code] = ExactMatchSkillExtractor(subontology.competency_framework)
        yield from self.skill_extractor_by_soc[soc_code].candidate_skills(source_object)



import random
from collections import Counter
from typing import List, Union, Iterator

import torch

Sentence = List["word"]
Doc = List[Sentence]

# sentences, but in index
Corpus = List[List[int]]

# Doc or Corpus, list of list of (idx or raw str)
Text = Union[Doc, Corpus]

def flatten_text(text: Text) -> Iterator[Union[int, str]]:
    """flatten text into one list"""
    for sent in text:
        for token in sent:
            yield token

def count_text(text: Text) -> Counter:
    """:return counter of each token in the text"""
    return Counter(
        flatten_text(text)
    )

class Vocab:
    def __init__(self, doc: Doc,
                 cutoff=10,
                 unk_label="<unk>",
                 reserved_tokens: List[str] = None):
        if reserved_tokens is None:
            reserved_tokens = []
        self.unk_label = unk_label

        # sort based on freq
        #   still reserve freq ≤ cutoff
        self.token_freq = count_text(doc)

        # in case doc contains unk_label symbol, not using enumerate
        #    (since index would discontinuous)
        #    get the list and update idx2token with list
        # most freq token to lower idx number
        #   unk index = 0, then reserved tokens, then others
        uniq_tokens = [unk_label] + reserved_tokens
        uniq_tokens += [token
                        for (token, ct) in self.token_freq.most_common()
                        if ct >= cutoff and token not in uniq_tokens]
        self.idx2token = {
            i: token for i, token in enumerate(uniq_tokens)
        }
        self.token2idx = dict(map(reversed, self.idx2token.items()))

    def __len__(self):
        """only the len of unique tokens (not counting freq ≤ cutoff)"""
        return len(self.idx2token)

    def __getitem__(self, token):
        """return the index of token, token str or iterable of str
            if not in token2idx (either freq ≤ cutoff or not in vocab)
                mapped to unk_label (0)"""
        if isinstance(token, (list, tuple)):
            return list(map(self.__getitem__, token))
        return self.token2idx.get(token, 0)

    def to_token(self, index):
        """return the token from indice, index int or iterable of int"""
        if isinstance(index, (list, tuple)):
            return list(map(self.to_token, index))
        return self.idx2token[index]

    def get_counts(self, token):
        """return the freq of token (still preserve freq ≤ cutoff),
         token str or iterable of str"""
        if isinstance(token, (list, tuple)):
            return list(map(self.token_freq.get, token))
        return self.token_freq[token]

    def lookup(self, doc: Doc) -> Doc:
        """map rare words to unk_label, if any"""
        # first map to idx (if rare or not exist map to 0)
        #      then get token
        return [[self.to_token(self[tk]) for tk in sent]
                for sent in doc]

    def doc2corpus(self, doc: Doc) -> Corpus:
        """map token (str) to corpus (idx)"""
        # first map rare then map to index
        return [self[sentence] for sentence in self.lookup(doc)]


def ngram(text: Text, n: int):
    text_flatten = list(flatten_text(text))
    return [text_flatten[i:i+n] for i in range(len(text_flatten)+1)]



import io
import os
import re
from typing import List, Optional
from zipfile import ZipFile

from nltk import word_tokenize, sent_tokenize

from torcherist.config import dataset_dir

Sentence = List["word"]

dataset_dir = os.path.join(dataset_dir, "text")

def read_ptb(filepath=os.path.join(dataset_dir, "ptb.zip")) -> List[Sentence]:
    """penn treebank, note that they already have <unk> unknown token"""
    with ZipFile(filepath) as ptb_zip:
        with ptb_zip.open("ptb/ptb.train.txt") as f:
            raw_text = io.TextIOWrapper(f).read()
    return [line.split() for line in raw_text.split("\n")]


def read_time_machine(filepath=os.path.join(dataset_dir, "timemachine.txt")) -> List[Sentence]:
    """also see dataloader.text.load_time_machine"""
    with open(filepath, 'r') as f:
        raw_text = f.read()

    doc = []
    for sent in sent_tokenize(raw_text):
        # append empty space
        tokens = [
            (token, " ")
             for token in word_tokenize(
             re.sub('[^A-Za-z]+', ' ', sent).strip().lower()
            )
        ]
        doc.extend([token for tuple in tokens for token in tuple])
    return doc

def read_nmt(filepath=os.path.join(dataset_dir, "fra-eng.zip"),
             num_examples: Optional[int] = None):
    """machine translation: english-french"""
    with ZipFile(filepath) as ptb_zip:
        with ptb_zip.open("fra-eng/fra.txt") as f:
            raw_text = io.TextIOWrapper(f).read()

    def no_space(char, prev_char):
        return char in set(',.!?') and prev_char != ' '

    text = raw_text.replace('\u202f', ' ').replace('\xa0', ' ').lower()
    text = ''.join([
        ' ' + char if i > 0 and no_space(char, text[i - 1]) else char
        for i, char in enumerate(text)
    ])

    # enize dataset
    source, target = [], []
    for i, line in enumerate(text.split('\n')):
        if num_examples and i > num_examples:
            break
        parts = line.split('\t')
        if len(parts) == 2:
            source.append(parts[0].split(' '))
            target.append(parts[1].split(' '))
    return source, target



from torcherist.dataloader.text import *
from torcherist.dataset.text import read_ptb
import pytest
import numpy as np
import random

def test_sequential_text_sampler():
    my_seq = np.arange(35)
    for X, y in sequential_text_sampler(my_seq, batch_size=2, m=5):
        X, y = X.numpy(), y.numpy()
        assert np.all(X + 1 == y)
    prev_seq_end = None
    for X, y in sequential_text_sampler(my_seq, by="sequential", batch_size=2, m = 5):
        X, y = X.numpy(), y.numpy()
        if prev_seq_end is not None:
            assert np.all(prev_seq_end + 1 == X[:,0])
        prev_seq_end = X[:,-1]
        assert np.all(X + 1 == y)

def test_time_machine():
    data_iter, vocab = load_time_machine(batch_size=32,
                                         m=35,
                                         by="random")
    # 26 en char & empty space & <unk>
    assert len(vocab) == 28
    for X, y in data_iter:
        # for each batch
        for XX, yy in zip(X, y):
            XX, yy = XX.numpy().flatten(), yy.numpy().flatten()
            assert len(XX) == len(yy)
            assert np.all(XX[1:] == yy[:-1])

    data_iter, vocab = load_time_machine(batch_size=32,
                                         m=35,
                                         by="sequential")
    # 26 en char & empty space & <unk>
    assert len(vocab) == 28
    for X, y in data_iter:
        # for each batch
        for XX, yy in zip(X, y):
            XX, yy = XX.numpy().flatten(), yy.numpy().flatten()
            assert len(XX) == len(yy)
            assert np.all(XX[1:] == yy[:-1])

    # test if data_iter can be iter multiple times
    for X, y in data_iter:
        # for each batch
        for XX, yy in zip(X, y):
            XX, yy = XX.numpy().flatten(), yy.numpy().flatten()
            assert len(XX) == len(yy)
            assert np.all(XX[1:] == yy[:-1])



# coding: utf-8
"""Use noun phrases with specific endings to extract skills from job postings"""

import logging

import nltk
try:
    nltk.pos_tag('test')
except LookupError:
    nltk.download('averaged_perceptron_tagger')

try:
    from nltk.tokenize.moses import MosesDetokenizer
except LookupError:
    nltk.download('perluniprops')
    from nltk.tokenize.moses import MosesDetokenizer

from .base import SkillExtractor, CandidateSkill, CandidateSkillYielder

from typing import Dict


def sentences_words_pos(document):
    """Chops raw text into part-of-speech (POS)-tagged words in sentences

    Args:
        document (string) A document in text format

    Returns: (list) of sentences, each being a list of word/POS pair

    Example:
        sentences_words_pos(
            '* Develop and maintain relationship with key members of ' +
            'ESPN\u2019s Spanish speaking editorial team'
        )
        [ # list of sentences
            [ # list of word/POS pairs
                ('*', 'NN'),
                ('Develop', 'NNP'),
                ('and', 'CC'),
                ('maintain', 'VB'),
                ('relationship', 'NN'),
                ('with', 'IN'),
                ('key', 'JJ'),
                ('members', 'NNS'),
                ('of', 'IN'),
                ('ESPN', 'NNP'),
                ('’', 'NNP'),
                ('s', 'VBD'),
                ('Spanish', 'JJ'),
                ('speaking', 'NN'),
                ('editorial', 'NN'),
                ('team', 'NN')
            ]
        ]
    """
    try:
        sentences = nltk.sent_tokenize(document.encode('utf-8'))
    except:
        sentences = nltk.sent_tokenize(document)
    sentences = [nltk.wordpunct_tokenize(sent) for sent in sentences]

    sentences = [nltk.pos_tag(sent) for sent in sentences]
    return sentences


def noun_phrases_in_line_with_context(line):
    """Generate noun phrases in the given line of text

    Args:
        text (string): A line of raw text

    Yields:
        tuples, each with two strings:
            - a noun phrase
            - the context of the noun phrase (currently defined as the surrounding sentence)
    """
    if line.strip() != "\n":
        output = sentences_words_pos(line)

        grammar = r"""
                  NP: {<JJ.*>*<NN.*><NN.*>*}   # chunk adjectives and nouns
                   """
        cp = nltk.RegexpParser(grammar)
        for sent in output:
            if sent:
                tree = cp.parse(sent)
                for subtree in tree.subtrees():
                    if subtree.label() == 'NP':
                        np = ""
                        for node in subtree:
                            try:
                                np += node[0].encode('utf-8') + ' '
                            except:
                                np += node[0] + ' '
                        np = np.strip()
                        logging.debug('Yielding noun phrase %s with context %s', np, sent)
                        yield np, sent


BULLETS = ['+', '*', '-']


def is_bulleted(string):
    """Whether or not a given string begins a 'bullet' character

    A bullet character is understood to indicate list membership.
    Differeing common bullet characters are checked.

    Args:
    string (string): Any string

    Returns: (bool) whether or not the string begins with one of the characters
        in a predefined list of common bullets
    """
    string_first = string.split(' ', 1)[0].replace('\n', '')
    try:
        if string_first[0] in BULLETS:
            return True
        else:
            return False
    except:
        return False


BEGINNING_SUBSTRINGS_TO_REMOVE = [
    "+ ",
    "â €¢ ",
    "Â · ",
    "Â · Â Â Â Â Â Â Â ",
    "\. ",
    "\- ",
    "/ ",
    "**,** ",
    "** ",
    "* ",
    "á ",
    "### ",
    "‰ Û¢ ",
    "å á å å å å å å å ",
    "å á ",
    "á "
]


def clean_beginning(string):
    """Clean the beginning of a string of common undesired formatting substrings

    Args:
    string (string): Any string

    Returns: The string with beginning formatting substrings removed
    """
    for substring in BEGINNING_SUBSTRINGS_TO_REMOVE:
        if string.startswith(substring):
            return string.replace(substring, "")
    return string


class NPEndPatternExtractor(SkillExtractor):
    """Identify noun phrases with certain ending words (e.g 'skills', 'abilities') as skills

    Args:
        endings (list): Single words that should identify the ending of a noun phrase
            as being a skill
        stop_phrases (list): Noun phrases that should not be considered skills
        only_bulleted_lines (bool, default True): Whether or not to only consider lines
            that look like they are items in a list
    """
    def __init__(self, endings, stop_phrases, only_bulleted_lines=True, confidence=95,  *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.endings = endings
        self.stop_phrases = stop_phrases
        self.only_bulleted_lines = only_bulleted_lines
        self.detokenizer = MosesDetokenizer()
        self.confidence = confidence

    @property
    def description(self):
        return f'Noun phrases ending with one of: {self.endings}\n'
        f'Excluding phrases: {self.stop_phrases}\n'
        f'Only using bulleted lines: {self.only_bulleted_lines}'

    def candidate_skills(self, source_object: Dict) -> CandidateSkillYielder:
        """Generate candidate skills from the source object

        Looks for all noun phrases with one of the configured endings.

        Args:
            source_object (dict) A single source object to extract skills from.

        Yields: all candidate skills (algorithms.skill_extractors.base.CandidateSkill)
            found in the job posting
        """
        document = self.transform_func(source_object)
        for cleaned_phrase, context, phrase_start in self.noun_phrases_matching_endings(document):
            orig_context = self.detokenizer.detokenize([t[0] for t in context], return_str=True)
            logging.info(
                'Yielding candidate skill %s in context %s',
                cleaned_phrase,
                orig_context
            )
            yield CandidateSkill(
                skill_name=cleaned_phrase,
                matched_skill_identifier=None,
                confidence=self.confidence,
                context=orig_context,
                start_index=phrase_start,
                document_id=source_object['id'],
                document_type=source_object['@type'],
                source_object=source_object,
                skill_extractor_name=self.name
            )

    def noun_phrases_matching_endings(self, document):
        """From the given document, generate noun phrases ending with one of the configured terms

        Args:
            document (string) A raw text document, such as a job posting

        Yields:
            tuples, each with:
                - a noun phrase
                - the context of the noun phrase (currently defined as the surrounding sentence)
                - the index of the start of the phrase in the document
        """
        lines = document.split('\n')
        phrase_start = 0
        for line in lines:
            if not self.only_bulleted_lines or is_bulleted(line):
                for noun_phrase, context in noun_phrases_in_line_with_context(line):
                    term_list = noun_phrase.split()
                    if term_list[-1].lower() in self.endings:
                        cleaned_phrase = clean_beginning(noun_phrase).lower()
                        if cleaned_phrase not in self.stop_phrases:
                            yield cleaned_phrase, context, phrase_start
                    phrase_start += len(noun_phrase)


class SkillEndingPatternExtractor(NPEndPatternExtractor):
    """Identify noun phrases ending with 'skill' or 'skills' as skills"""

    GENERIC_SKILL_PHRASES = set([
        'advanced skills',
        'basic skills',
        'bonus skills',
        'career skills',
        'core skills',
        'demonstrable skill',
        'demonstrated skill',
        'demonstrated skills',
        'desired additional skills',
        'desired skills',
        'encouraged skills',
        'essential skills',
        'essential job skills',
        'following requisite skills',
        'following skill',
        'following skills',
        'individual skills',
        'job skills',
        'key skills',
        'new skills',
        'or skill',
        'other skills',
        'personal skills',
        'preferred skills',
        'primary skills',
        'professional skills',
        'related skills',
        'required job skills',
        'required skills',
        'skill',
        'skills',
        'special skills',
        'specific skills',
        'welcomed skill',
    ])

    def __init__(self, *args, **kwargs):
        super().__init__(
            endings=['skills', 'skill'],
            stop_phrases=self.GENERIC_SKILL_PHRASES,
            *args,
            **kwargs
        )

    @property
    def name(self):
        bulleted_str = '_bulleted' if self.only_bulleted_lines else ''
        return f'noun_phrase_skill{bulleted_str}'


class AbilityEndingPatternExtractor(NPEndPatternExtractor):
    """Identify noun phrases ending in 'ability' or 'abilities' as skills"""
    def __init__(self, *args, **kwargs):
        super().__init__(
            endings=['ability', 'abilities'],
            stop_phrases=['demonstrated ability', 'ability'],
            *args,
            **kwargs
        )

    @property
    def name(self):
        bulleted_str = '_bulleted' if self.only_bulleted_lines else ''
        return f'noun_phrase_ability{bulleted_str}'



from typing import Callable, Text, List
from collections import MutableMapping, KeysView
import json
from statistics import median, mean
from functools import total_ordering
import itertools
import requests


def research_hub_url(name: Text) -> Text:
    """Retrieve the current Research Hub URL for ontologies"""
    return 'https://open-skills-datasets.s3-us-west-2.amazonaws.com/v3/ontologies/' + name + '.json'


@total_ordering
class Competency(object):
    """Represents a competency not necessarily tied to an ontology

    Args:
        identifier: A unique identifier for this competency. Choose the identifier wisely as it will be used for equivalence with other competency objects
        name: A name for the competency (e.g. Microsoft Office)
        categories: Optional text categories for the competency that is not a higher-level competency itself
    """

    def __init__(self, identifier: Text, name: Text=None, categories: List[Text]=None, **kwargs):
        self.identifier = identifier
        self.name = name or ''
        self.categories = categories or []
        self.other_attributes = kwargs
        self.children = set()
        self.parents = set()

    @classmethod
    def from_jsonld(cls, jsonld_input):
        extra_kwargs = dict((key, jsonld_input[key]) for key in jsonld_input.keys() if key not in {'@type', '@id', 'name', 'competencyCategory', 'hasChild', 'isChildOf'})
        obj = cls(
            identifier=jsonld_input['@id'],
            name=jsonld_input.get('name', ''),
            categories=jsonld_input.get('competencyCategory', None),
            **extra_kwargs
        )
        for jsonld_child_obj in jsonld_input.get('hasChild', []):
            obj.add_child(cls(identifier=jsonld_child_obj['@id']))
        for jsonld_parent_obj in jsonld_input.get('isChildOf', []):
            obj.add_parent(cls(identifier=jsonld_parent_obj['@id']))
        return obj

    def __eq__(self, other):
        if not isinstance(self, other.__class__):
            return NotImplemented
        return self.identifier == other.identifier

    def __lt__(self, other):
        if not isinstance(self, other.__class__):
            return NotImplemented
        return self.identifier < other.identifier

    def __hash__(self):
        return hash(self.identifier)

    def __repr__(self):
        return f'Competency(identifier={self.identifier}, name={self.name}, categories={self.categories}, {self.other_attributes})'

    @property
    def jsonld_id(self):
        return {
            '@type': 'Competency',
            '@id': self.identifier
        }

    @property
    def jsonld_full(self):
        attributes = {
            '@type': 'Competency',
            '@id': self.identifier,
            'name': self.name,
            'competencyCategory': self.categories,
            'hasChild': [child.jsonld_id for child in self.children],
            'isChildOf': [parent.jsonld_id for parent in self.parents],
        }
        for key, value in self.other_attributes.items():
            attributes[key] = value
        return attributes

    def add_child(self, child):
        if not isinstance(child, Competency):
            raise ValueError('All children of a Competency must be Competencies themselves')
        if child not in self.children:
            self.children.add(child)
            child.add_parent(self)

    def add_parent(self, parent):
        if not isinstance(parent, Competency):
            raise ValueError('All parents of a Competency must be Competencies themselves')
        if parent not in self.parents:
            self.parents.add(parent)
            parent.add_child(self)


@total_ordering
class Occupation(object):
    """Represents an occupation that may or may not be part of an ontology

    Args:
        identifier: A unique identifier for this occupation. Choose the identifier wisely as it will be used for equivalence with other occupation objects
        name: A name for the occupation (e.g. Civil Engineer)
    """

    def __init__(self, identifier, name=None, **kwargs):
        self.identifier = identifier
        self.name = name or ''
        self.other_attributes = kwargs
        self.children = set()
        self.parents = set()

    @classmethod
    def from_jsonld(cls, jsonld_input):
        extra_kwargs = dict((key, jsonld_input[key]) for key in jsonld_input.keys() if key not in {'@type', '@id', 'name', 'hasChild', 'isChildOf'})
        obj = cls(
            identifier=jsonld_input['@id'],
            name=jsonld_input.get('name', ''),
            **extra_kwargs
        )
        for jsonld_child_obj in jsonld_input.get('hasChild', []):
            obj.add_child(cls(identifier=jsonld_child_obj['@id']))
        for jsonld_parent_obj in jsonld_input.get('isChildOf', []):
            obj.add_parent(cls(identifier=jsonld_parent_obj['@id']))
        return obj

    def __eq__(self, other):
        if not isinstance(self, other.__class__):
            return NotImplemented
        return self.identifier == other.identifier

    def __lt__(self, other):
        if not isinstance(self, other.__class__):
            return NotImplemented
        return self.identifier < other.identifier

    def __hash__(self):
        return hash(self.identifier)

    def __repr__(self):
        return f'Occupation(identifier={self.identifier}, name={self.name}, {self.other_attributes})'

    @property
    def jsonld_id(self):
        return {
            '@type': 'Occupation',
            '@id': self.identifier
        }

    @property
    def jsonld_full(self):
        attributes = {
            '@type': 'Occupation',
            '@id': self.identifier,
            'name': self.name,
            'hasChild': [child.jsonld_id for child in self.children],
            'isChildOf': [parent.jsonld_id for parent in self.parents],
        }
        for key, value in self.other_attributes.items():
            attributes[key] = value
        return attributes


    def add_child(self, child):
        if not isinstance(child, Occupation):
            raise ValueError('All children of a Occupation must be Occupations themselves')
        if child not in self.children:
            self.children.add(child)
            child.add_parent(self)

    def add_parent(self, parent):
        if not isinstance(parent, Occupation):
            raise ValueError('All parents of a Occupation must be Occupations themselves')
        if parent not in self.parents:
            self.parents.add(parent)
            parent.add_child(self)


class DummyOccupation(Occupation):
    def __init__(self):
        super().__init__('0', '')

class DummyCompetency(Competency):
    def __init__(self):
        super().__init__('0', '')


class CompetencyOccupationEdge(object):
    def __init__(self, competency, occupation, identifier=None):
        self.competency = competency
        self.occupation = occupation
        self.identifier = f'competency={competency.identifier}&occupation={occupation.identifier}'
        if identifier:
            assert identifier == self.identifier

    def __repr__(self):
        return f'CompetencyOccupationEdge(competency={self.competency}, occupation={self.occupation})'

    @classmethod
    def from_jsonld(cls, jsonld_input):
        obj = cls(
            identifier=jsonld_input['@id'],
            competency=Competency.from_jsonld(jsonld_input['competency']),
            occupation=Occupation.from_jsonld(jsonld_input['occupation'])
        )
        return obj

    @property
    def jsonld_id(self):
        return {
            '@type': 'CompetencyOccupationEdge',
            '@id': self.identifier
        }

    @property
    def jsonld_full(self):
        return {
            '@type': 'CompetencyOccupationEdge',
            '@id': self.identifier,
            'competency': self.competency.jsonld_id,
            'occupation': self.occupation.jsonld_id
        }

    def __eq__(self, other):
        if isinstance(self, other.__class__):
            return self.competency == other.competency \
                   and self.occupation == other.occupation

    def __hash__(self):
        return hash(self.competency) + hash(self.occupation)


class CompetencyFramework(MutableMapping):
    """A list of competencies and metadata about them

    Implements MutableMapping, so the competencies may be interacted with as a dictionary.
    """

    def __init__(self, name=None, description=None, competencies=None):
        self.name = name or ''
        self.description = description or ''
        self._competencies = dict()
        if competencies:
            self.competencies = competencies

    def __setitem__(self, key, value):
        self._competencies[key] = value

    def __getitem__(self, key):
        return self._competencies[key]

    def __delitem__(self, key):
        del self._competencies[key]

    def __iter__(self):
        return iter(self._competencies)

    def __len__(self):
        return len(self._competencies)

    def add(self, value):
        if value.identifier in self._competencies:
            raise ValueError(f"{value} already in framework")
        self._competencies[value.identifier] = value

    @property
    def competencies(self):
        return self._competencies

    @competencies.setter
    def competencies(self, competencies):
        for competency in competencies:
            self._competencies[competency.identifier] = competency


class CompetencyOntology(object):
    """An ontology of competencies and occupations (each referred to as nodes) and the edges between them

    Can be initialized with:
        - a JSON-LD string, in which case all nodes and edges will be initialized from the parsed JSON-LD
        - a URL, in which case the URL is presumed to contain JSON-LD, parsed, and all nodes and edges will be initialized from the parsed JSON-LD
        - a research hub name, in which case the Open Skills Research Hub is presumed to host an ontology by that name. a URL is built based on the name and the nodes and edges will be initialized from the parsed JSON-LD
        - a set of edges, in which case all nodes will be initialized with any that are present in the edge list
    """
    name = 'unnamed_ontology'

    def __init__(
            self,
            edges=None,
            name=None,
            competency_name=None,
            competency_description=None,
            research_hub_name=None,
            url=None,
            jsonld_string=None
    ):
        self.name = name
        self._initialize_empty(competency_name, competency_description)
        if jsonld_string:
            self._build_from_jsonld(jsonld_string)
        elif url:
            self._build_from_url(url)
        elif research_hub_name:
            self._build_from_research_hub(research_hub_name)
        elif edges:
            self._build_from_edges(edges)

    def _initialize_empty(self, competency_framework_name, competency_framework_description):
        self.competency_framework = CompetencyFramework(
            name=competency_framework_name,
            description=competency_framework_description,
        )
        self._occupations = dict()
        self._competency_occupation_edges = set()

    def _build_from_research_hub(self, name):
        self._build_from_url(research_hub_url(name))

    def _build_from_url(self, url):
        jsonld_string = requests.get(url).content
        self._build_from_jsonld(jsonld_string)

    def _build_from_jsonld(self, jsonld_string: Text):
        jsonld_input = json.loads(jsonld_string)
        self.name = jsonld_input.get('name', 'unnamed ontology')
        for competency_jsonld in jsonld_input['competencies']:
            self.add_competency(Competency.from_jsonld(competency_jsonld))
        for occupation_jsonld in jsonld_input['occupations']:
            self.add_occupation(Occupation.from_jsonld(occupation_jsonld))
        for edge_jsonld in jsonld_input['edges']:
            self.add_edge(edge=CompetencyOccupationEdge.from_jsonld(edge_jsonld))

    def _build_from_edges(self, edges):
        self._competency_occupation_edges = edges
        self.competency_framework.competencies = [edge.competency for edge in edges]
        self._occupations = dict((edge.occupation.identifier, edge.occupation) for edge in edges)

    def __eq__(self, other):
        if isinstance(self, other.__class__):
            return \
                self.competencies == other.competencies \
                and self.occupations == other.occupations \
                and self.edges == other.edges


    def __str__(self):
        return f'Competency Framework with {len(self.competencies)} competencies' + \
               f', {len(self.occupations)} occupations' + \
               f', {len(self.edges)} competency-occupation edges'

    @property
    def competencies(self):
        return set(self.competency_framework.values()) - {DummyCompetency()}

    @property
    def occupations(self):
        return set(self._occupations.values()) - {DummyOccupation()}

    @property
    def edges(self):
        values = set([
            edge for edge in self._competency_occupation_edges
            if edge.occupation != DummyOccupation() and edge.competency != DummyCompetency()
        ])
        return values

    def add_competency(self, competency: Competency):
        if not isinstance(competency, Competency):
            raise ValueError('Must add competency objects')
        if competency.identifier not in self.competency_framework:
            self.competency_framework[competency.identifier] = competency
            self.add_edge(competency=competency, occupation=DummyOccupation())
        else:
            self.competency_framework[competency.identifier].children.update(competency.children)
            self.competency_framework[competency.identifier].parents.update(competency.parents)

    def add_occupation(self, occupation: Occupation):
        if not isinstance(occupation, Occupation):
            raise ValueError('Must add occupation objects')
        if occupation.identifier not in self._occupations:
            self._occupations[occupation.identifier] = occupation
            self.add_edge(competency=DummyCompetency(), occupation=occupation)
        else:
            self._occupations[occupation.identifier].children.update(occupation.children)
            self._occupations[occupation.identifier].parents.update(occupation.parents)

    def add_edge(self, occupation: Occupation=None, competency: Competency=None, edge: CompetencyOccupationEdge=None):
        if edge:
            self._competency_occupation_edges.add(edge)
            return

        if not isinstance(occupation, Occupation) or not isinstance(competency, Competency):
            raise ValueError('Must pass both an occupation and competency')
        self.add_competency(competency)
        self.add_occupation(occupation)
        edge = CompetencyOccupationEdge(
            occupation=self._occupations[occupation.identifier],
            competency=self.competency_framework[competency.identifier]
        )
        if edge not in self._competency_occupation_edges:
            self._competency_occupation_edges.add(edge)

    def filter_by(self, func: Callable, competency_name=None, competency_description=None):
        """Produce an ontology that is filtered by a callable that takes in an edge
        """
        matching_edges = set(edge for edge in self._competency_occupation_edges if func(edge))
        return CompetencyOntology(edges=matching_edges, competency_name=competency_name, competency_description=competency_description)

    @property
    def jsonld(self):
        return json.dumps({
            'name': self.name,
            'competencies': [
                competency.jsonld_full
                for competency in
                sorted(self.competencies, key=lambda competency: competency.identifier)
            ],
            'occupations': [
                occupation.jsonld_full
                for occupation in
                sorted(self.occupations, key=lambda occupation: occupation.identifier)
            ],
            'edges': [
                edge.jsonld_full
                for edge in
                sorted(self.edges, key=lambda edge: edge.identifier)
            ]
        }, sort_keys=True)

    def save(self, storage):
        storage.write(self.jsonld.encode('utf-8'), self.name + '.json')

    @property
    def occupation_counts_per_competency(self):
        counts = []
        for competency, edges in itertools.groupby(
                sorted(self._competency_occupation_edges, key=lambda edge: edge.competency),
                lambda edge: edge.competency
        ):
            if competency != DummyCompetency():
                counts.append(len(set([
                    edge.occupation for edge in list(edges) if edge.occupation != DummyOccupation()
                ])))
        return counts

    @property
    def competency_counts_per_occupation(self):
        counts = []
        for occupation, edges in itertools.groupby(
                sorted(self._competency_occupation_edges, key=lambda edge: edge.occupation),
                lambda edge: edge.occupation
        ):
            if occupation != DummyOccupation():
                counts.append(len(set([
                    edge.competency for edge in list(edges)
                    if edge.competency != DummyCompetency()
                ])))
        return counts

    def print_summary_stats(self):
        print(f'Ontology summary statistics for {self.name}')
        print(f'Num competencies: {len(self.competency_framework)}')
        print(f'Num occupations: {len(self.occupations)}')
        print(f'Num competency-occupation edges: {len(self.edges)}')
        print(f'Median occupations per competency: {median(self.occupation_counts_per_competency)}')
        print(f'Median competencies per occupation: {median(self.competency_counts_per_occupation)}')
        print(f'Mean occupations per competency: {mean(self.occupation_counts_per_competency)}')
        print(f'Mean competencies per occupation: {mean(self.competency_counts_per_occupation)}')
